﻿Imports UPLib

Public Class MainForm
    Dim Level As Package
    Dim LevelModel As Model
    Dim LevelFile As String

    Private Sub MainForm_Shown(sender As System.Object, e As System.EventArgs) Handles MyBase.Shown
        My.Computer.FileSystem.CurrentDirectory = My.Application.Info.DirectoryPath

        Try
            If IO.File.Exists("BSPTool.ini") Then
                Using r As New IO.StreamReader("BSPTool.ini")
                    While Not r.EndOfStream
                        Dim line = r.ReadLine
                        If line.StartsWith("RemoveVoid=", StringComparison.OrdinalIgnoreCase) Then
                            cbRemoveVoid.Checked = line.Substring(line.IndexOf("=") + 1)
                        ElseIf line.StartsWith("FixInvisibleSemisolids=", StringComparison.OrdinalIgnoreCase) Then
                            cbFixInvisibleSemisolids.Checked = line.Substring(line.IndexOf("=") + 1)
                        ElseIf line.StartsWith("FixMoreInvisibleSemisolids=", StringComparison.OrdinalIgnoreCase) Then
                            cbFixMoreInvisibleSemisolids.Checked = line.Substring(line.IndexOf("=") + 1)
                        ElseIf line.StartsWith("FixRenderBounds=", StringComparison.OrdinalIgnoreCase) Then
                            cbFixRenderBounds.Checked = line.Substring(line.IndexOf("=") + 1)
                        ElseIf line.StartsWith("EnlargeRenderBounds=", StringComparison.OrdinalIgnoreCase) Then
                            cbEnlargeRenderBounds.Checked = line.Substring(line.IndexOf("=") + 1)
                        ElseIf line.StartsWith("RemoveLevelSummary=", StringComparison.OrdinalIgnoreCase) Then
                            cbRemoveLevelSummary.Checked = line.Substring(line.IndexOf("=") + 1)
                        End If
                    End While
                End Using
            End If
        Catch
        End Try

        If My.Application.CommandLineArgs.Count = 1 AndAlso IO.File.Exists(My.Application.CommandLineArgs(0)) Then
            LoadMap(My.Application.CommandLineArgs(0))
        End If

    End Sub

    Private Sub MainForm_FormClosing(sender As System.Object, e As System.Windows.Forms.FormClosingEventArgs) Handles MyBase.FormClosing
        Try
            IO.File.WriteAllText("BSPTool.ini",
                                 "RemoveVoid=" & cbRemoveVoid.Checked & vbCrLf &
                                 "FixInvisibleSemisolids=" & cbFixInvisibleSemisolids.Checked & vbCrLf &
                                 "FixMoreInvisibleSemisolids=" & cbFixMoreInvisibleSemisolids.Checked & vbCrLf &
                                 "FixRenderBounds=" & cbFixRenderBounds.Checked & vbCrLf &
                                 "EnlargeRenderBounds=" & cbEnlargeRenderBounds.Checked & vbCrLf &
                                 "RemoveLevelSummary=" & cbRemoveLevelSummary.Checked)
        Catch
        End Try
    End Sub

    Private Sub LoadMap(file As String)
        Cursor = Cursors.WaitCursor
        LevelFile = file
        Level = New Package(file)
        LevelModel = GetModel(FindLevelModel(Level))
        SaveDlg.InitialDirectory = IO.Path.GetDirectoryName(file)
        SaveDlg.FileName = IO.Path.GetFileName(file)
        btnOpenMap.Font = New Font(btnOpenMap.Font.FontFamily, 8)
        btnOpenMap.Text = file
        ActionsPanel.Visible = True
        Cursor = Cursors.Default
    End Sub

    Private Sub MainForm_DragEnter(sender As System.Object, e As System.Windows.Forms.DragEventArgs) Handles MyBase.DragEnter
        If e.Data.GetDataPresent(DataFormats.FileDrop) Then
            e.Effect = DragDropEffects.All
        End If
    End Sub

    Private Sub MainForm_DragDrop(sender As System.Object, e As System.Windows.Forms.DragEventArgs) Handles MyBase.DragDrop
        If e.Data.GetDataPresent(DataFormats.FileDrop) Then
            Dim files As String() = e.Data.GetData(DataFormats.FileDrop)
            LoadMap(files(0))
        End If
    End Sub

    Private Sub llAbout_LinkClicked(sender As System.Object, e As System.Windows.Forms.LinkLabelLinkClickedEventArgs) Handles llAbout.LinkClicked
        MessageBox.Show(My.Resources.AboutText, "About")
    End Sub

    Private Sub btnOpenMap_Click(sender As System.Object, e As System.EventArgs) Handles btnOpenMap.Click
        If OpenDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            LoadMap(OpenDlg.FileName)
        End If
    End Sub

    Private Sub btnExploreBSP_Click(sender As System.Object, e As System.EventArgs) Handles btnExploreBSP.Click
        ExploreForm.LevelModel = LevelModel
        ExploreForm.LevelFile = LevelFile
        ExploreForm.ShowDialog()
    End Sub

    Private Shared Function CalcRenderBounds(model As Model, iNode As Integer, Optional dontSet As Boolean = False) As BoundingBox
        Dim curBound As BoundingBox

        Dim node = model.Nodes(iNode)

        If node.iChild0 <> -1 Then curBound.Extend(CalcRenderBounds(model, node.iChild0))
        If node.iChild1 <> -1 Then curBound.Extend(CalcRenderBounds(model, node.iChild1))
        If node.iChild2 <> -1 Then curBound.Extend(CalcRenderBounds(model, node.iChild2, True))

        For i = 0 To node.NumVertices - 1
            curBound.Extend(model.Points(model.Verts(node.iVertPool + i).pVertex))
        Next

        If Not dontSet Then
            If node.iRenderBound <> -1 Then
                Dim origBound = model.Bounds(node.iRenderBound)
                origBound.Extend(curBound)
                model.Bounds(node.iRenderBound) = origBound
            End If
        End If

        Return curBound
    End Function

    Private Shared Sub FixBrownPolys(model As Model, Optional iNode As Integer = 0)
        Dim node = model.Nodes(iNode)

        If node.iChild0 <> -1 Then FixBrownPolys(model, node.iChild0)
        If node.iChild1 <> -1 Then FixBrownPolys(model, node.iChild1)

        Dim goodZone = node.iZone1
        If goodZone = 0 Then goodZone = node.iZone0
        While node.iChild2 <> -1
            iNode = node.iChild2
            node = model.Nodes(iNode)

            If node.iZone1 = 0 Then
                node.iZone1 = goodZone
                model.Nodes(iNode) = node
            End If

        End While
    End Sub

    Enum EIsBrushSemiSolid
        Unknown
        No
        Yes
    End Enum

    Dim _IsBrushSemiSolid As EIsBrushSemiSolid()

    Private Function IsBrushSemiSolid(objRef As Integer) As Boolean
        If _IsBrushSemiSolid(objRef - 1) <> EIsBrushSemiSolid.Unknown Then
            Return _IsBrushSemiSolid(objRef - 1) = EIsBrushSemiSolid.Yes
        Else
            Dim props = Level.ExportTable(objRef - 1).GetProperties()
            Dim flags = props.GetInt("PolyFlags")
            If flags And EPolyFlags.PF_Semisolid Then
                _IsBrushSemiSolid(objRef - 1) = EIsBrushSemiSolid.Yes
                Return True
            Else
                _IsBrushSemiSolid(objRef - 1) = EIsBrushSemiSolid.No
                Return False
            End If
        End If
    End Function

    Private Sub FixNonbrownPolys(model As Model, Optional iNode As Integer = 0)
        Dim node = model.Nodes(iNode)

        If node.iChild0 <> -1 Then FixNonbrownPolys(model, node.iChild0)
        If node.iChild1 <> -1 Then FixNonbrownPolys(model, node.iChild1)

        Dim origNode = node
        While node.iChild2 <> -1
            iNode = node.iChild2
            node = model.Nodes(iNode)

            If node.NumVertices >= 3 Then

                If node.iSurf <> -1 AndAlso model.Surfs(node.iSurf).Actor <> 0 AndAlso IsBrushSemiSolid(model.Surfs(node.iSurf).Actor) Then

                    Dim avgVerts As New Vector(0, 0, 0) 'What the fuck??? Why does this initialize to non-zero?
                    For i = 0 To node.NumVertices - 1
                        Dim pt = model.Points(model.Verts(node.iVertPool + i).pVertex)
                        avgVerts += pt
                    Next
                    avgVerts /= node.NumVertices
                    Dim pointToSample = avgVerts + New Vector(node.Plane.X, node.Plane.Y, node.Plane.Z)

                    Dim iNewZone = model.GetLocZone(pointToSample).ZoneNumber

                    If iNewZone <> 0 Then
                        node.iZone1 = iNewZone
                        model.Nodes(iNode) = node
                    Else
                        'MessageBox.Show("Anomaly at " & pointToSample.ToPropertyString, "", MessageBoxButtons.OK, MessageBoxIcon.Warning)
                    End If

                End If

            End If
        End While
    End Sub

    Private Sub btnSaveMap_Click(sender As System.Object, e As System.EventArgs) Handles btnSaveMap.Click
        If SaveDlg.ShowDialog = Windows.Forms.DialogResult.OK Then

            Cursor = Cursors.WaitCursor

            If cbFixInvisibleSemisolids.Checked Then ' Takes care of most such errors.
                FixBrownPolys(LevelModel)

                For i As Integer = 0 To LevelModel.Nodes.Count - 1
                    Dim node = LevelModel.Nodes(i)
                    If node.iZone1 = 0 Then node.iZone1 = node.iZone0
                    LevelModel.Nodes(i) = node
                Next
            End If

            If cbFixMoreInvisibleSemisolids.Checked Then ' Fix polys assigned to wrong, but non-void zone. More error prone.
                ReDim _IsBrushSemiSolid(Level.ExportTable.Count - 1)
                FixNonbrownPolys(LevelModel)
                ReDim _IsBrushSemiSolid(-1)
            End If

            If cbFixRenderBounds.Checked Then
                CalcRenderBounds(LevelModel, 0)
            End If

            If cbEnlargeRenderBounds.Checked Then
                For i = 0 To LevelModel.Bounds.Count - 1
                    Dim bound = LevelModel.Bounds(i)
                    Dim center = (bound.Max + bound.Min) / 2
                    Dim radius = (bound.Max - bound.Min) / 2

                    radius *= 2

                    bound.Min = center - radius
                    bound.Max = center + radius

                    bound.Min.X = Math.Min(32768, Math.Max(-32768, bound.Min.X))
                    bound.Min.Y = Math.Min(32768, Math.Max(-32768, bound.Min.Y))
                    bound.Min.Z = Math.Min(32768, Math.Max(-32768, bound.Min.Z))
                    bound.Max.X = Math.Min(32768, Math.Max(-32768, bound.Max.X))
                    bound.Max.Y = Math.Min(32768, Math.Max(-32768, bound.Max.Y))
                    bound.Max.Z = Math.Min(32768, Math.Max(-32768, bound.Max.Z))

                    LevelModel.Bounds(i) = bound
                Next
            End If

            If cbRemoveVoid.Checked Then
                For i = 0 To LevelModel.Nodes.Count - 1
                    Dim node = LevelModel.Nodes(i)
                    If node.iZone0 = 0 Then node.iZone0 = 63
                    If node.iZone1 = 0 Then node.iZone1 = 63
                    LevelModel.Nodes(i) = node
                Next
            End If

            SetModel(FindLevelModel(Level), LevelModel)

            If cbRemoveLevelSummary.Checked Then
                'Dim levelInfo = level.ExportTable.Find(Function(x) NameEq(x.ClassNameStr, "LevelInfo"))
                'If levelInfo IsNot Nothing Then
                '    Dim props = levelInfo.GetProperties
                '    Dim iSummaryProp = props.FindProp("Summary")
                '    If iSummaryProp <> -1 Then
                '        props.Props.RemoveAt(iSummaryProp)
                '        levelInfo.OriginalOffset = 0
                '        levelInfo.OriginalSize = 0
                '        levelInfo.Data.Clear()
                '        levelInfo.Data.W.WriteProperties(props)
                '        levelInfo.ObjectFlags = levelInfo.ObjectFlags And Not EObjectFlags.RF_HasStack
                '    End If
                'End If
                'Dim levelSummary = level.ExportTable.Find(Function(x) NameEq(x.ClassNameStr, "LevelSummary"))
                'If levelSummary IsNot Nothing Then levelSummary.NullOut()
                '' to/do: import table

                Dim NAME_LevelSummary = Level.FindName("LevelSummary")
                If NAME_LevelSummary <> -1 Then Level.NameTable(NAME_LevelSummary).Name = "SpecialEvent"
            End If

            Level.Save(SaveDlg.FileName)

            Cursor = Cursors.Default

            Close()
        End If
    End Sub
End Class